#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include "cublas_v2.h"

#include <thrust/extrema.h>
#include <thrust/device_ptr.h>
#include <thrust/device_vector.h>

#include "time.h"
#include "windows.h"
#include "float.h"
#include "math.h"
#include <iostream>
using namespace std;
#include <stdio.h>


double CPUtime=0;
double GPUthrust_time=0;
double GPUcublas_time=0;
double GPUkernel_time=0;
int times=0;
#define ONLY_reduction_kernel
#define test1

#ifdef test1
	#define code_length   504
#endif

#ifdef test2
#define code_length   1008
#endif

#ifdef test3
#define code_length   2000
#endif

#ifdef test4
#define code_length   4000
#endif

#ifdef test5
#define code_length   8000
#endif

#ifdef test6
#define code_length   10000
#endif

#ifdef test7
	#define code_length   504
#endif

#ifdef test8
#define code_length   1008
#endif

#ifdef test9
#define code_length   2000
#endif

#ifdef test10
#define code_length   4000
#endif

#ifdef test11
#define code_length   8000
#endif

#ifdef test12
#define code_length   10000
#endif

//#define code_length   10000
//#define win_num       code_length/20




void read_code_exam(FILE *fp, int &n, int &m, int *&ir, int *&jc){
    // The parity-check matrix H is an nxm sparse matrix
    fscanf(fp, "%d", &n);// number of v-nodes
    fscanf(fp, "%d", &m);// number of c-nodes
    
    int maxVdeg, maxCdeg, temp;
    fscanf(fp, "%d", &maxVdeg);// maximum degree of v-nodes
    fscanf(fp, "%d", &maxCdeg);// maximum degree of c-nodes
    
    // jc[i]: accumulated degrees of v-nodes
    jc = new int[n+1]; jc[0] = 0;
    for(int i=0; i<n; i++){
        fscanf(fp, "%d", &temp);// number of 1's in the i-th col of H
        jc[i+1] = jc[i] + temp;
    }
    for(int j=0; j<m; j++)	fscanf(fp, "%d", &temp);// number of 1's in the j-th row of H
    
    // ir[k]: the index of the check node that is connected with the k-th edge
    ir = new int[jc[n]];
    for(int i=0,k=0; i<n; i++){// i: index of v-nodes; k: index of edges numbered from v-nodes' viewpoint
        for(int j=0; j<maxVdeg; j++){
            fscanf(fp, "%d", &temp);
            if(temp){
                ir[k++] = temp - 1;// the k-th edge connects the i-th v-node with the ir[k]-th c-node
            }
        }
    }
}

int beliefPropagation_exam(int *ir, int *jc, int m, int n, char *sid, char *syn, char *rec, char *bmem, double *dmem, double pb, double pc){
    // ellr: LLRs of edge; cllr: LLRs of c-node; vllr: LLRs of v-node. vllr and cllr share the same bulk of memory
    char *ellr_sgn = bmem, *cllr_sgn = ellr_sgn+jc[n];
    double *vllrin = dmem, *ellr = vllrin+n, *cllr = ellr+jc[n], *vllr = cllr;
    
    //initialize variable-to-check messages
    for(int i=0; i<n; i++){
        vllrin[i] = (1-2*sid[i])*log((1-pc)/pc) + log((1-pb)/pb);
        for(int k=jc[i]; k<jc[i+1]; k++){
            ellr[k] = vllrin[i];
        }
    }
    
    for(int iteration=0, sameCount=0; iteration<100; iteration++){
        //Step 1: compute check-to-variable messages
        memset(cllr, 0, m*sizeof(double));
        memcpy(cllr_sgn, syn, m);
        for(int k=0; k<jc[n]; k++){
            ellr_sgn[k] = (ellr[k]<0);
            ellr[k] = -log(tanh(max(fabs(ellr[k]), DBL_MIN)/2));
            cllr_sgn[ir[k]] ^= ellr_sgn[k];
            cllr[ir[k]] += ellr[k]; 
        }
        for(int k=0; k<jc[n]; k++){
            ellr_sgn[k] = cllr_sgn[ir[k]]^ellr_sgn[k];
            ellr[k] = -log(tanh(max(cllr[ir[k]]-ellr[k], DBL_MIN)/2));
            ellr[k] *= (1-2*ellr_sgn[k]);
        }
        
        //Step 2: compute variable-to-check messages
        int ndiffs = 0;
        memcpy(vllr, vllrin, n*sizeof(double));
        for(int i=0; i<n; i++){
            for(int k=jc[i]; k<jc[i+1]; k++)    vllr[i] += ellr[k];
            char temp = (vllr[i]<0);
            ndiffs += (rec[i]^temp);
            rec[i] = temp;
            for(int k=jc[i]; k<jc[i+1]; k++)    ellr[k] = vllr[i] - ellr[k];
        }
        
        //Step 3: test convergence and syndrome condition
        if(ndiffs){// different
            sameCount = 0;
            memcpy(cllr_sgn, syn, m);
            for(int i=0; i<n; i++){
                for(int k=jc[i]; k<jc[i+1]; k++){
                    cllr_sgn[ir[k]] ^= rec[i];
                }
            }
            for(int j=0; j<m; j++){
                if(cllr_sgn[j])
                    break;
                else if(j==m-1)
                    return 1; //all syndrome checks satisfied
                else;
            }
        }else{// the same
            if(sameCount==4) return 0; //convergence (to wrong answer)
            else sameCount++;
        }
    }
    return 0;
}

void read_code(FILE *fp, int &n, int &m, int* &ir, int* &jc){
    // The parity-check matrix H is an nxm sparse matrix
    fscanf(fp, "%d", &n);// number of v-nodes
    fscanf(fp, "%d", &m);// number of c-nodes
    
    int maxCdeg, temp;
    fscanf(fp, "%d", &maxCdeg);// maximum degree of c-nodes
    jc = new int[m*maxCdeg];// jc[k]: the index of the v-node that is connected with the k-th edge
    ir = new int[m+1]; ir[0] = 0;// ir[j]: accumulated degrees of c-nodes
    for(int j=0,k=0; j<m; j++){// j: index of c-nodes; k: index of edges numbered from c-nodes' viewpoint
        ir[j+1] = ir[j];
        for(int i=0; i<maxCdeg; i++){
            fscanf(fp, "%d", &temp);
            if(temp){
                jc[k++] = temp - 1;// the k-th edge connects the j-th c-node with the jc[k]-th v-node
                ir[j+1]++;
            }
        }
    }
}

//#ifdef SWBP

float findbestw(float *bias, float *gamma, int n, int w)
{
    int k, l, u = (w>>1);
    float sigma2 = 0.0, sum = 0.0;
    
    // node 0
    int truncw = u;
    for(l=1; l<=u; l++)	sum += gamma[l];
    bias[0] = sum/truncw;  
    sigma2 = (gamma[0]-bias[0])*(gamma[0]-bias[0]);
    
    // nodes 1 to n-1
    for(k=1; k<n; k++){
        sum -= gamma[k];
        sum += gamma[k-1];
        if(k<=u){// nodes 1 to u
            sum += gamma[k+u];
            truncw++;       
        }else if((k+u)<n){// nodes (u+1) to (n-1-u)
            sum -= gamma[(k-1)-u];
            sum += gamma[k+u];
        }else{// nodes (n-u) to (n-1)
            sum -= gamma[(k-1)-u];
            truncw--;
        }
        bias[k] = min(max(sum/truncw, 0.000001), 0.5);
        sigma2 += (gamma[k]-bias[k])*(gamma[k]-bias[k]);        
    }
    return sigma2;
}
//#endif SWBP
// input: r: over all belief; n: length of code; w: window size
// output: sigma2: square of sigma
__global__ void GPU_findbest(float *sigma2, double *bias, double *gamma, int n, int w)
{
	int k, l, u;
	float s2=0.0,sum=0.0;

	int i = threadIdx.x+blockIdx.x*blockDim.x;
	
	
	//u= (w>>1);
	u=(w+i*2)>>1;
	//printf("in %d thread, u=%d\n", i, u);
	 // node 0
    int truncw = u;
    for(l=1; l<=u; l++)	sum += gamma[l];
    bias[0] = sum/truncw;  
    s2 = (gamma[0]-bias[0])*(gamma[0]-bias[0]);
    
    // nodes 1 to n-1
    for(k=1; k<n; k++){
        sum -= gamma[k];
        sum += gamma[k-1];
        if(k<=u){// nodes 1 to u
            sum += gamma[k+u];
            truncw++;       
        }else if((k+u)<n){// nodes (u+1) to (n-1-u)
            sum -= gamma[(k-1)-u];
            sum += gamma[k+u];
        }else{// nodes (n-u) to (n-1)
            sum -= gamma[(k-1)-u];
            truncw--;
        }
        bias[k] = min(max(sum/truncw, 0.000001), 0.5);
        s2 += (gamma[k]-bias[k])*(gamma[k]-bias[k]);        
    }

	sigma2[i]=s2;	
//	printf("in %d thread, u=%d, sigma2=%f\n", i, u, s2);
	
	
}

__global__ void GPU_findbest1(float *sigma2, float *gamma, int n, int w)
{
	int k, l, u;
	float s2=0.0,sum=0.0, bias;
	int i = threadIdx.x+blockIdx.x*blockDim.x;
		
	//u= (w>>1);
	u=(w+i*2)>>1;
	//printf("in %d thread, u=%d\n", i, u);
	 // node 0
    int truncw = u;
    for(l=1; l<=u; l++)	sum += gamma[l];
	
	bias = sum/truncw;  
    s2 = (gamma[0]-bias)*(gamma[0]-bias);
    
    // nodes 1 to n-1
    for(k=1; k<n; k++){
        sum -= gamma[k];
        sum += gamma[k-1];
        if(k<=u){// nodes 1 to u
            sum += gamma[k+u];
            truncw++;       
        }else if((k+u)<n){// nodes (u+1) to (n-1-u)
            sum -= gamma[(k-1)-u];
            sum += gamma[k+u];
        }else{// nodes (n-u) to (n-1)
            sum -= gamma[(k-1)-u];
            truncw--;
        }
        bias = min(max(sum/truncw, 0.000001), 0.5);
        s2 += (gamma[k]-bias)*(gamma[k]-bias);        
    }

	sigma2[i]=s2;	
//	printf("in %d thread, u=%d, sigma2=%f\n", i, u, s2);
	
	
}

//#define DSIZE code_length
// nTPB should be a power-of-2
#define nTPB 256
#define MAX_KERNEL_BLOCKS 30
#define MAX_BLOCKS ((code_length/nTPB)+1)
#define MIN(a,b) ((a>b)?b:a)
//#define FLOAT_MIN -1.0f
#define FLOAT_MAX 1.0e30
__device__ volatile float blk_vals[MAX_BLOCKS];
__device__ volatile int   blk_idxs[MAX_BLOCKS];
__device__ int   blk_num = 0;

//template <typename T>
#define T float
__global__ void min_idx_kernel(const T *data, const int dsize, int *result){

  __shared__ volatile T   vals[nTPB];
  __shared__ volatile int idxs[nTPB];
  __shared__ volatile int last_block;
  int idx = threadIdx.x+blockDim.x*blockIdx.x;
  last_block = 0;
  T   my_val = FLOAT_MAX;
  //T   my_val = FLOAT_MIN;
  int my_idx = -1;
  // sweep from global memory
  while (idx < dsize){
    if (data[idx] < my_val) {my_val = data[idx]; my_idx = idx;}
	//if (data[idx] > my_val) {my_val = data[idx]; my_idx = idx;}
    idx += blockDim.x*gridDim.x;}
  // populate shared memory
  vals[threadIdx.x] = my_val;
  idxs[threadIdx.x] = my_idx;
  __syncthreads();
  // sweep in shared memory
  for (int i = (nTPB>>1); i > 0; i>>=1){
    if (threadIdx.x < i)
      //if (vals[threadIdx.x] < vals[threadIdx.x + i]) {vals[threadIdx.x] = vals[threadIdx.x+i]; idxs[threadIdx.x] = idxs[threadIdx.x+i]; }
	  if (vals[threadIdx.x] > vals[threadIdx.x + i]) {vals[threadIdx.x] = vals[threadIdx.x+i]; idxs[threadIdx.x] = idxs[threadIdx.x+i]; }
    __syncthreads();}
 // *result = idxs[0];
  // perform block-level reduction
  if (!threadIdx.x){
    blk_vals[blockIdx.x] = vals[0];
    blk_idxs[blockIdx.x] = idxs[0];
    if (atomicAdd(&blk_num, 1) == gridDim.x - 1) // then I am the last block
      last_block = 1;}
  __syncthreads();
  if (last_block){
    idx = threadIdx.x;
    my_val = FLOAT_MAX;
	//my_val = FLOAT_MIN;
    my_idx = -1;
    while (idx < gridDim.x){
    if (blk_vals[idx] < my_val) {my_val = blk_vals[idx]; my_idx = blk_idxs[idx]; }
	//if (blk_vals[idx] > my_val) {my_val = blk_vals[idx]; my_idx = blk_idxs[idx]; }
      idx += blockDim.x;}
  // populate shared memory
    vals[threadIdx.x] = my_val;
    idxs[threadIdx.x] = my_idx;
    __syncthreads();
  // sweep in shared memory
    for (int i = (nTPB>>1); i > 0; i>>=1){
      if (threadIdx.x < i)
        //if (vals[threadIdx.x] < vals[threadIdx.x + i]) {vals[threadIdx.x] = vals[threadIdx.x+i]; idxs[threadIdx.x] = idxs[threadIdx.x+i]; }
		if (vals[threadIdx.x] > vals[threadIdx.x + i]) {vals[threadIdx.x] = vals[threadIdx.x+i]; idxs[threadIdx.x] = idxs[threadIdx.x+i]; }
      __syncthreads();}
    if (!threadIdx.x)
      *result = idxs[0];
    }
}

int beliefPropagation(int *ir, int *jc, int m, int n, char *sid, char *syn, char *rec, char *bmem, double *dmem, double pb, double pc){
    // ellr: LLRs of edges; cllr: LLRs of c-nodes; vllr: LLRs of v-nodes. // vllr and cllr share the same bulk of memory
    char *ellr_sgn = bmem, *cllr_sgn = ellr_sgn+ir[m];
    double *vllrin = dmem, *ellr = vllrin+n, *cllr = ellr+ir[m], *vllr = cllr;
    
    //initialize variable-to-check messages
    for(int i=0; i<n; i++){
        vllrin[i] = (1-2*sid[i])*log((1-pc)/pc) + log((1-pb)/pb);
    }
    
    for(int k=0; k<ir[m]; k++){// for each edge
        ellr[k] = vllrin[jc[k]];
    }
    
    for(int iteration=0, sameCount=0; iteration<50; iteration++){
        //Step 1: compute check-to-variable messages
        for(int k=0; k<ir[m]; k++){// for each edge
            ellr_sgn[k] = (ellr[k]<0);
            ellr[k] = -log(tanh(max(fabs(ellr[k]), DBL_MIN)/2));
        }
        
        memset(cllr, 0, m*sizeof(double));
        memcpy(cllr_sgn, syn, m);
        for (int j=0; j<m; j++) {// for each syndrome bit
            for (int k=ir[j]; k<ir[j+1]; k++) {
                cllr_sgn[j] ^= ellr_sgn[k];
                cllr[j] += ellr[k];
            }
            for (int k=ir[j]; k<ir[j+1]; k++) {
                ellr_sgn[k] ^= cllr_sgn[j];
                ellr[k] = -log(tanh(max(cllr[j]-ellr[k], DBL_MIN)/2));
                ellr[k] *= (1-2*ellr_sgn[k]);
            }
        }
        
        //Step 2: compute variable-to-check messages
        int ndiffs = 0;
        memcpy(vllr, vllrin, n*sizeof(double));// vllr <= vllrin
        for(int k=0; k<ir[m]; k++){// for each edge
            vllr[jc[k]] += ellr[k];
        }
        for(int k=0; k<ir[m]; k++){// for each edge
            ellr[k] = vllr[jc[k]] - ellr[k];
        }
        for(int i=0; i<n; i++){
            char temp = (vllr[i]<0);
            ndiffs += (rec[i]^temp);
            rec[i] = temp;
        }
        
        //Step 3: test convergence and syndrome condition
        if(ndiffs){// different
            sameCount = 0;
            memcpy(cllr_sgn, syn, m);
            for (int j=0; j<m; j++) {// for each syndrome bit
                for (int k=ir[j]; k<ir[j+1]; k++) {
                    cllr_sgn[j] ^= rec[jc[k]];
                }
            }
            for(int j=0; j<m; j++){
                if(cllr_sgn[j])
                    break;
                else if(j==m-1)
                    return 1; //all syndrome checks satisfied
                else;
            }
        }else{// the same
            if(sameCount==4) return 0; //convergence (to wrong answer)
            else sameCount++;
        }

//#ifdef SWBP
		int k,l;
		int w,bestw, u, truncw, wsize=0;
		float *overall = (float*)malloc(n*sizeof(float));
		//double overall[code_length] ={0};
		float *bias= (float*)malloc(n*sizeof(float));
		float cost, mincost;
		float sum=0.0, prob;

		
		
		if((iteration+1)%10==0)
		{
			for(k=0;k<n;k++)
			{
				overall[k]=(1.0/(1.0+exp(vllr[k])));
			}
			if(wsize>0)
			{//fix-win-size
				bestw=wsize;
			}
			else
			{  
				
				LARGE_INTEGER  num;  
				long start,end,freq;  
				QueryPerformanceFrequency(&num);  
				freq=num.QuadPart;  
				QueryPerformanceCounter(&num);   
				start= num.QuadPart;   	
				
				//adaptive-win-size
				bestw=21;
				mincost=findbestw(bias,overall, n, bestw);
				for(w=41; w<min(n, 10001); w+=2)
				{
					cost =findbestw(bias, overall, n, w);
					if(cost<mincost)
					{
						bestw=w;
						mincost=cost;
					}
				}

				
				QueryPerformanceCounter(&num);   
				end= num.QuadPart;    
				printf("CPUϴִʱΪtime=%f(ms)\n",(double)(end-start)*1000/freq); 
				CPUtime+=(double)(end-start)*1000/freq;
				//GPU adaptive win size
				                //[win_num];

			
			

				float	*dev_sigma2;
				float	*dev_bias,
						*dev_overall;
				

				#ifdef test1
				#define Nb   25
				#define Nt   25
				#endif

				#ifdef test2
				#define Nb   32
				#define Nt   32
				#endif

				#ifdef test3
				#define Nb   45
				#define Nt   45
				#endif

				#ifdef test4
				#define Nb   64
				#define Nt   64
				#endif

				#ifdef test5
				#define Nb   90
				#define Nt   90
				#endif

				#ifdef test6
				#define Nb   100
				#define Nt   100
				#endif

				#ifdef test7
				#define Nb   25
				#define Nt   25
				#endif

				#ifdef test8
				#define Nb   32
				#define Nt   32
				#endif

				#ifdef test9
				#define Nb   45
				#define Nt   45
				#endif

				#ifdef test10
				#define Nb   64
				#define Nt   64
				#endif

				#ifdef test11
				#define Nb   90
				#define Nt   90
				#endif

				#ifdef test12
				#define Nb   100
				#define Nt   100
				#endif
				float sigma2[Nb*Nt];

				
				cublasStatus_t status;
				cublasHandle_t handle;
				status = cublasCreate(&handle);
				if (status != CUBLAS_STATUS_SUCCESS) 
				{ printf ("CUBLAS initialization failed\n"); return 0; }


				cudaError_t cudaStatus = cudaMalloc((void**)&dev_sigma2, Nb*Nt* sizeof(float));
				if (cudaStatus != cudaSuccess) {
					fprintf(stderr, "cudaMalloc failed!");
					return 0;
				}

				cudaStatus = cudaMalloc((void**)&dev_bias, n * sizeof(float));
				if (cudaStatus != cudaSuccess) {
					fprintf(stderr, "cudaMalloc failed!");
					return 0;
				}

				cudaStatus = cudaMalloc((void**)&dev_overall, n * sizeof(float));
				if (cudaStatus != cudaSuccess) {
					fprintf(stderr, "cudaMalloc failed!");
					return 0;
				}


				cudaStatus = cudaMemcpy(dev_overall, overall, n * sizeof(float), cudaMemcpyHostToDevice);
				if (cudaStatus != cudaSuccess) {
					fprintf(stderr, "cudaMemcpy failed!");
					return 0;
				}

				 cudaStatus = cudaSetDevice(0);
				 if (cudaStatus != cudaSuccess) {
					 fprintf(stderr, "cudaSetDevice failed!  Do you have a CUDA-capable GPU installed?");
					 return 0;
				 }
				 //__global__ void GPU_findbest(int *sigma2, double r, int n, int w)

				float time_elapsed=0;
				cudaEvent_t gstart,gstop;
				cudaEventCreate(&gstart);    //Event
				cudaEventCreate(&gstop);
				cudaEventRecord( gstart,0);    //¼ǰʱ			
			
				// Launch a kernel on the GPU with one thread for each element.
				 
				 w=1;
				  // GPU_findbest<<<Nb, Nt>>>(dev_sigma2, dev_bias, dev_overall, n, w);
				GPU_findbest1<<<Nb, Nt>>>(dev_sigma2, dev_overall, n, w);


				cudaEventRecord( gstop,0);    //¼ǰʱ 
				cudaEventSynchronize(gstart);    //Waits for an event to complete.
				cudaEventSynchronize(gstop);    //Waits for an event to complete.Record֮ǰ
				cudaEventElapsedTime(&time_elapsed,gstart,gstop);    //ʱ
			//	cudaEventDestroy(gstart);    //destory the event
			//	cudaEventDestroy(gstop);
				
				//printf("GPUϲִʱΪtime=%f(ms)\n",time_elapsed);
				

				    // Check for any errors launching the kernel
				cudaStatus = cudaGetLastError();
				if (cudaStatus != cudaSuccess) {
					 fprintf(stderr, "addKernel launch failed: %s\n", cudaGetErrorString(cudaStatus));
        
				}
				   
				// cudaDeviceSynchronize waits for the kernel to finish, and returns
				// any errors encountered during the launch.
				cudaStatus = cudaDeviceSynchronize();
				if (cudaStatus != cudaSuccess) {
				  fprintf(stderr, "cudaDeviceSynchronize returned error code %d after launching addKernel!\n", cudaStatus);
				    return 0;
				}

				// Copy output vector from GPU buffer to host memory.
				cudaStatus = cudaMemcpy(sigma2, dev_sigma2, Nb*Nt * sizeof(float), cudaMemcpyDeviceToHost);
				if (cudaStatus != cudaSuccess) {
				    fprintf(stderr, "cudaMemcpy failed!");
				    return 0;
				}
				cudaFree(dev_sigma2);
				cudaFree(dev_bias);
				cudaFree(dev_overall);

				
				int bestu;
//				int incx=1;
				
			
				/*double *dev_sigma2;*/
				cudaStatus = cudaMalloc((void**)&dev_sigma2, Nb*Nt* sizeof(float));
				if (cudaStatus != cudaSuccess) {
					fprintf(stderr, "cudaMalloc failed!");
					return 0;
				}

				cudaStatus = cudaMemcpy(dev_sigma2, sigma2, Nb* Nt * sizeof(float), cudaMemcpyHostToDevice);
				if (cudaStatus != cudaSuccess) {
					fprintf(stderr, "cudaMemcpy failed!");
					return 0;
				}
#ifndef ONLY_reduction_kernel
				//thrust algorithm
				cudaEventRecord( gstart,0);    //¼ǰʱ

				thrust::device_ptr<float> d_ptr = thrust::device_pointer_cast(dev_sigma2);
				thrust::device_vector<float>::iterator d_it = thrust::min_element(d_ptr, d_ptr + Nb*Nt);
				bestu = d_it - (thrust::device_vector<float>::iterator)d_ptr;
				//cublasIsamin (handle, Nb*Nt, dev_sigma2,  incx, &bestu);				

				cudaEventRecord( gstop,0);    //¼ǰʱ 
				cudaEventSynchronize(gstart);    //Waits for an event to complete.
				cudaEventSynchronize(gstop);    //Waits for an event to complete.Record֮ǰ
				float time_thrust=0;
				cudaEventElapsedTime(&time_thrust,gstart,gstop);    //ʱ
				//cudaEventDestroy(gstart);    //destory the event
				//cudaEventDestroy(gstop);
				time_thrust+=time_elapsed;
				GPUthrust_time+=time_thrust;
				printf("GPUϲthrust time=%f(ms)\n",time_thrust);



				//clublasIsamin algoritm
				cudaEventRecord( gstart,0);    //¼ǰʱ

				cublasIsamin (handle, Nb*Nt, dev_sigma2,  incx, &bestu);
				

				cudaEventRecord( gstop,0);    //¼ǰʱ 
				cudaEventSynchronize(gstart);    //Waits for an event to complete.
				cudaEventSynchronize(gstop);    //Waits for an event to complete.Record֮ǰ
				float time_cublas=0;
				cudaEventElapsedTime(&time_cublas,gstart,gstop);    //ʱ
				
				time_cublas+=time_elapsed;
				GPUcublas_time+=time_cublas;
				printf("GPUϲcublasmin time=%f(ms)\n",time_cublas);

#endif				

				int *d_min_index;
				cudaMalloc(&d_min_index, sizeof(int));
				cudaEventRecord( gstart,0);    //¼ǰʱ

				min_idx_kernel<<<MIN(MAX_KERNEL_BLOCKS, ((code_length+nTPB-1)/nTPB)), nTPB>>>(dev_sigma2, Nb*Nt, d_min_index);
				
				cudaEventRecord( gstop,0);    //¼ǰʱ
				cudaEventSynchronize(gstart);    //Waits for an event to complete.
				cudaEventSynchronize(gstop);    //Waits for an event to complete.Record֮ǰ
				float time_kernel=0;
				cudaEventElapsedTime(&time_kernel,gstart,gstop);    //ʱ
				cudaEventDestroy(gstart);    //destory the event
				cudaEventDestroy(gstop);
				time_kernel+=time_elapsed;
				GPUkernel_time+=time_kernel; times++;
				printf("GPUϲkernel time=%f(ms)\n",time_kernel);

				cudaMemcpy(&bestu, d_min_index, sizeof(int), cudaMemcpyDeviceToHost);


				cudaFree(dev_sigma2);			
				cublasDestroy(handle);

				   
			

			}

			//refine vllr
			u = (bestw>>1);
                for(k=0; k<n; k++){
                    if(k==0){// the 0th node
                        sum = 0.0;
                        for(l=1; l<=u; l++)	sum += overall[l];	
                        truncw = u;
                    }else{
                        sum -= overall[k];
                        sum += overall[k-1];
                        if(k<=u){// nodes 1 to u
                            sum += overall[k+u];
                            truncw++;
                        }else if((k+u)<n){// nodes (u+1) to (n-u-1)
                            sum -= overall[(k-1)-u];
                            sum += overall[k+u];
                        }else{// nodes (n-u) to (n-1)
                            sum -= overall[(k-1)-u];
                            truncw--;
                        }
                    }
                    prob = min(max(sum/truncw, 0.000001), 0.5);
                    vllr[k] = log((1-prob)/prob);
                }
			
		}
//#endif
    }
    return 0;
}

int main(int argc, const char* argv[])
{
    int n, m, *ir=NULL, *jc=NULL;
	//FILE *fp = fopen("PEGirReg252x504.txt", "rt");
	//FILE *fp = fopen("PEGReg5Kx10K", "rt");
#ifdef test1
	FILE *fp = fopen("Reg504252.dat", "rt");
#endif test1

#ifdef test2
	FILE *fp = fopen("Reg1008504.dat", "rt");
#endif test2

#ifdef test3
	FILE *fp = fopen("Reg2k1k.dat", "rt");
#endif test3

#ifdef test4
	FILE *fp = fopen("Reg4k2k.dat", "rt");
#endif test4

#ifdef test5
	FILE *fp = fopen("Reg8k4k.dat", "rt");
#endif test5

#ifdef test6
	FILE *fp = fopen("Reg10k5k.dat", "rt");
#endif test6

#ifdef test7
	FILE *fp = fopen("irReg504252.dat", "rt");
#endif 

#ifdef test8
	FILE *fp = fopen("irReg1008504.dat", "rt");
#endif 

#ifdef test9
	FILE *fp = fopen("irReg2k1k.dat", "rt");
#endif 

#ifdef test10
	FILE *fp = fopen("irReg4k2k.dat", "rt");
#endif 

#ifdef test11
	FILE *fp = fopen("irReg8k4k.dat", "rt");
#endif 

#ifdef test12
	FILE *fp = fopen("irReg10k5k.dat", "rt");
#endif 

	if(!fp){
        cout<<"LDPC code file doesn't exist!"<<endl;
        return 1;
    }else{
        //read_code_exam(fp, n, m, ir, jc);
		read_code(fp, n, m, ir, jc);
        fclose(fp);
    }
    
	char *src = new char[3*n+m], *rec = src+n, *sid = rec+n, *syn = sid+n;
    char *bmem = new char[ir[m]+m];
    double *dmem = new double[ir[m]+2*n];
    
    double pb = 0.5, pc = 0.08;
    double fer = 0.0, ber = 0.0;
    int ntests = 200;
    
    clock_t start = clock();
    for(int l=0; l<ntests; l++){
        // step 1: generate source and side information
        for(int i=0; i<n; i++){
            src[i] = (rand()<(RAND_MAX*pb));
            sid[i] = (src[i]^(rand()<(RAND_MAX*pc)));
        }

        // step 2: encoding via s = xH
        memset(syn, 0, m);
//		for(int i=0; i<n; i++){// for each v-node
//			for(int k=ir[i]; k<ir[i+1]; k++){// for each edge
//				syn[jc[k]] ^= src[i];
//			}
//		}
        
        for(int j=0; j<m; j++){// for each c-node
            for(int k=ir[j]; k<ir[j+1]; k++){// for each edge
                syn[j] ^= src[jc[k]];
            }
        }
        
        // step 3: decoding
		//beliefPropagation_exam(ir, jc, m, n, sid, syn, rec, bmem, dmem, pb, pc);
		beliefPropagation(ir, jc, m, n, sid, syn, rec, bmem, dmem, pb, pc);
        
        // step 4: correctness test
		int nerrs = 0;
		for(int i=0; i<n; i++)  nerrs += (rec[i]^src[i]);
        ber += nerrs;
        fer += (nerrs>0);
	}
    clock_t end = clock();
    cout<<"Running time: "<<(double)(end-start)/CLOCKS_PER_SEC<<"s"<<endl;
    
    ber /= (ntests*n);
    fer /= ntests;
    cout<<"FER = "<<fer<<", "<<"BER = "<<ber<<endl;

	CPUtime/=times;
	GPUthrust_time/=times;
	GPUcublas_time/=times;
	GPUkernel_time/=times;
	cout<<"average CPU time = "<<CPUtime<<", average GPU thrust time = "<<GPUthrust_time<<", average GPU cublas time = "<<GPUcublas_time<<", average GPU kernel time = "<<GPUkernel_time<<endl;

    delete[] jc;
    free(ir);
    delete[] src;
    delete[] bmem;
    delete[] dmem;
    return 0;
}