#include <cuda_runtime.h>
#include "device_launch_parameters.h"
//#include <helper_cuda.h>
//#include <helper_functions.h>

#include <iostream>
using namespace std;
#include <stdio.h>
/*	 
#include "reduction1.cuh"
#include "reduction1_template.cuh"
#include "reduction2.cuh"
#include "reduction2_template.cuh"
#include "reduction3.cuh"
#include "reduction3_template.cuh"
*/
// Two steps reduction


__global__ void reduction1_kernel(int *out, const int *in, size_t N)
{
    // lenght = threads (BlockDim.x)
    extern __shared__ int sPartials[];
    int sum = 0;
    const int tid = threadIdx.x;
    for (size_t i = blockIdx.x * blockDim.x + tid; i < N; i += blockDim.x * gridDim.x)
    {
        sum += in[i];
    }
    sPartials[tid] = sum;
    __syncthreads();

    unsigned int floowPow2 = blockDim.x;
    if (floowPow2 & (floowPow2 - 1))
    {
        while(floowPow2 & (floowPow2 - 1))
        {
            floowPow2 &= (floowPow2 - 1);
        }
        if (tid >= floowPow2)
        {
            sPartials[tid - floowPow2] += sPartials[tid];
        }
        __syncthreads();
    }

    for (int activeTrheads = floowPow2 / 2; activeTrheads > 32; activeTrheads /= 2)
    {
        if (tid < activeTrheads)
        {
            sPartials[tid] += sPartials[tid + activeTrheads];
        }
        __syncthreads();
    }

    if (tid < 32)
    {
        volatile int *wsSum = sPartials;
        if (floowPow2 > 32)
        {
            wsSum[tid] += wsSum[tid + 32];
        }

        if (floowPow2 > 16) wsSum[tid] += wsSum[tid + 16];
        if (floowPow2 > 8) wsSum[tid] += wsSum[tid + 8];
        if (floowPow2 > 4) wsSum[tid] += wsSum[tid + 4];
        if (floowPow2 > 2) wsSum[tid] += wsSum[tid + 2];
        if (floowPow2 > 1) wsSum[tid] += wsSum[tid + 1];

        if (tid == 0)
        {
            volatile int *wsSum = sPartials;
            out[blockIdx.x] = wsSum[0];
        }
    }
}

void reduction1(int *answer, int *partial, const int *in, const size_t N, const int numBlocks, int numThreads)
{
    unsigned int sharedSize = numThreads * sizeof(int);

    // kernel execution
    reduction1_kernel<<<numBlocks, numThreads, sharedSize>>>(partial, in, N);
    reduction1_kernel<<<1, numThreads, sharedSize>>>(answer, partial, numBlocks);
}




typedef void (*pfnReduction)(int*, int*, const int*, size_t, int, int);

#define NUM_COUNT (1024 * 1024 * 1)
#define NUM_THREADS (1024)
 
double run(int cIterations, int *answer, int *partial, const int *in, const size_t N, 
    const int numBlocks, int numThreads, pfnReduction func)
{
    cudaEvent_t start, stop;
 //   checkCudaErrors( cudaEventCreate(&start) );
	cudaEventCreate(&start);
 //   checkCudaErrors( cudaEventCreate(&stop) );
	cudaEventCreate(&stop) ;
    cudaEventRecord(start);

    for (int i = 0; i < cIterations; ++i)
    {
        func(answer, partial, in, N, numBlocks, numThreads);
    }

    cudaEventRecord(stop);
//    checkCudaErrors( cudaThreadSynchronize() );
	cudaThreadSynchronize();
    float time = 0;
    cudaEventElapsedTime(&time, start, stop);

    return time;
}

int main(int argc, char const *argv[])
{
    int blocks = (NUM_COUNT + NUM_THREADS - 1) / NUM_THREADS;
    std::cout << "blocks: " << blocks << " threads: " << NUM_THREADS << std::endl;

    size_t nums_mem_size = NUM_COUNT * sizeof(int);
    int *h_nums = (int *)malloc(nums_mem_size);
    int *d_nums;
    //checkCudaErrors(cudaMalloc(&d_nums, nums_mem_size));
	cudaMalloc(&d_nums, nums_mem_size);

    size_t partical_mem_size = blocks * sizeof(int);
    int *d_partial;
    //checkCudaErrors(cudaMalloc(&d_partial, partical_mem_size));
	cudaMalloc(&d_partial, partical_mem_size);

    size_t answer_mem_size = 1 * sizeof(int);
    int *h_answer = (int *)malloc(answer_mem_size);
    int *d_answer;
    //checkCudaErrors(cudaMalloc(&d_answer, answer_mem_size));
	cudaMalloc(&d_answer, answer_mem_size);

    for (int i = 0; i < NUM_COUNT; ++i)
    {
        h_nums[i] = 1;
    }

    int host_answer = 0;
    for (int i = 0; i < NUM_COUNT; ++i)
    {
        host_answer += h_nums[i];
    }

    struct
    {
        std::string name;
        pfnReduction func;
    } rgTests[] = {
        { "simple loop", reduction1 },
        /*{ "simple loop template", reduction1t },
        { "atomicAdd", reduction2 },
        { "atomicAdd template", reduction2t },
        { "single pass", reduction3 },
        { "single pass template", reduction3t },
        */
    };

    int numTests = sizeof(rgTests) / sizeof(rgTests[0]);

    // memory copy
    //checkCudaErrors(cudaMemcpy(d_nums, h_nums, nums_mem_size, cudaMemcpyHostToDevice));
	cudaMemcpy(d_nums, h_nums, nums_mem_size, cudaMemcpyHostToDevice);

    for (int i = 0; i < numTests; ++i)
    {
        double time = run(100, d_answer, d_partial, 
                          d_nums, NUM_COUNT, blocks, 
                          NUM_THREADS, rgTests[i].func);

        //checkCudaErrors(cudaMemcpy(h_answer, d_answer, answer_mem_size, cudaMemcpyDeviceToHost));
		cudaMemcpy(h_answer, d_answer, answer_mem_size, cudaMemcpyDeviceToHost);
        
        std::string equal = (host_answer == *h_answer) ? "=" : "!=";

        std::cout << "simple loop" <<  " time: " << time  << "ms host answer (" << host_answer << ") "<< " device answer (" << *h_answer << ")"   << std::endl;
    } 
	 
    return 0;
}